@@ -15,13 +15,14 @@ import sys
 import copy
 from collections import OrderedDict
 from os import path, environ
+from gettext import gettext
 
 import yaml
 from jsonschema import validate
 
-from mockintosh import configs
+from mockintosh.constants import PROGRAM
 from mockintosh.exceptions import UnrecognizedConfigFileFormat
-from mockintosh.handlers import Request, Response  # noqa: F401
+from mockintosh.replicas import Request, Response  # noqa: F401
 from mockintosh.methods import _detect_engine, _nostderr, _import_from
 from mockintosh.recognizers import (
     PathRecognizer,
@@ -34,10 +35,11 @@ from mockintosh.recognizers import (
 from mockintosh.servers import HttpServer, TornadoImpl
 from mockintosh.performance import PerformanceProfile
 
-__version__ = "0.7"
+__version__ = "0.8"
 __location__ = path.abspath(path.dirname(__file__))
 
 should_cov = environ.get('COVERAGE_PROCESS_START', False)
+cov_no_run = environ.get('COVERAGE_NO_RUN', False)
 
 
 class Definition():
@@ -48,10 +50,7 @@ class Definition():
         self.source_dir = path.dirname(path.abspath(source)) if source is not None and is_file else None
         self.data = None
         self.schema = schema
-        if self.source is None:
-            self.data = configs.get_default()
-        else:
-            self.load()
+        self.load()
         self.orig_data = copy.deepcopy(self.data)
         self.validate()
         for service in self.data['services']:
@@ -175,6 +174,13 @@ class Definition():
         return service
 
 
+class CustomArgumentParser(argparse.ArgumentParser):
+    def error(self, message):
+        self.print_help(sys.stderr)
+        args = {'prog': self.prog, 'message': message}
+        self.exit(2, gettext('\n%(prog)s: error: %(message)s\n') % args)
+
+
 def get_schema():
     schema = None
     schema_path = path.join(__location__, 'schema.json')
@@ -212,7 +218,7 @@ def run(source, is_file=True, debug=False, interceptors=(), address='', services
             address=address,
             services_list=services_list
         )
-    except Exception:
+    except Exception:  # pragma: no cover
         logging.exception('Mock server loading error:')
         with _nostderr():
             raise
@@ -221,7 +227,7 @@ def run(source, is_file=True, debug=False, interceptors=(), address='', services
 
 def gracefully_exit(num, frame):
     atexit._run_exitfuncs()
-    if should_cov:
+    if should_cov:  # pragma: no cover
         sys.exit()
 
 
@@ -229,11 +235,11 @@ def cov_exit(cov):
     if should_cov:
         logging.debug('Stopping coverage')
         cov.stop()
-        cov.save()
+        cov.save()  # pragma: no cover
 
 
 def initiate():
-    if should_cov:
+    if should_cov:  # pragma: no cover
         signal.signal(signal.SIGTERM, gracefully_exit)
         logging.debug('Starting coverage')
         from coverage import Coverage
@@ -251,12 +257,12 @@ def initiate():
     This method parses the command-line arguments and handles the top-level initiations accordingly.
     """
 
-    ap = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
+    ap = CustomArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
     ap.add_argument(
         'source',
         help='Path to configuration file and (optional) a list of the service names\n'
              'to specify the services to be listened.',
-        nargs='*'
+        nargs='+'
     )
     ap.add_argument('-q', '--quiet', help='Less logging messages, only warnings and errors', action='store_true')
     ap.add_argument('-v', '--verbose', help='More logging messages, including debug', action='store_true')
@@ -288,16 +294,14 @@ def initiate():
         handler.setFormatter(logging.Formatter(fmt))
         logging.getLogger('').addHandler(handler)
 
-    logging.info("Mockintosh v%s is starting..." % __version__)
+    logging.info("%s v%s is starting..." % (PROGRAM.capitalize(), __version__))
 
     debug_mode = environ.get('DEBUG', False) or environ.get('MOCKINTOSH_DEBUG', False)
     if debug_mode:
         logging.debug('Tornado Web Server\'s debug mode is enabled!')
 
-    source = None
-    services_list = []
-    if len(args['source']) > 0:
-        source = args['source'][0]
-        services_list = args['source'][1:]
+    source = args['source'][0]
+    services_list = args['source'][1:]
 
-    run(source, debug=debug_mode, interceptors=interceptors, address=address, services_list=services_list)
+    if not cov_no_run:  # pragma: no cover
+        run(source, debug=debug_mode, interceptors=interceptors, address=address, services_list=services_list)
