@@ -14,8 +14,8 @@ JINJA = 'Jinja2'
 SHORT_PYBARS = 'hbs'
 SHORT_JINJA = 'j2'
 
-SUPPORTED_ENGINES = (PYBARS, JINJA)
-
 SPECIAL_CONTEXT = '%s_special_context' % PROGRAM
 
 JINJA_VARNAME_DICT = 'jinja_varname_dict'
+
+BASE64 = 'base64'
