@@ -10,28 +10,29 @@ import json
 import logging
 import os
 import re
-import urllib
 import socket
 import struct
-import time
 import traceback
+from urllib.parse import quote_plus
+from datetime import datetime, timezone
 from typing import (
     Union,
     Optional,
     Awaitable
 )
 
+import httpx
 import jsonschema
 import tornado.web
-from tornado.concurrent import Future
 from accept_types import parse_header
+from tornado.concurrent import Future
 
 import mockintosh
-from mockintosh.constants import PROGRAM, SUPPORTED_ENGINES, PYBARS, JINJA, SPECIAL_CONTEXT
-from mockintosh.exceptions import UnsupportedTemplateEngine
+from mockintosh.constants import PROGRAM, PYBARS, JINJA, SPECIAL_CONTEXT, BASE64
+from mockintosh.replicas import Request, Response
 from mockintosh.hbs.methods import Random as hbs_Random, Date as hbs_Date
 from mockintosh.j2.methods import Random as j2_Random, Date as j2_Date
-from mockintosh.methods import _safe_path_split, _detect_engine, _decoder
+from mockintosh.methods import _safe_path_split, _detect_engine, _b64encode
 from mockintosh.params import (
     PathParam,
     HeaderParam,
@@ -40,8 +41,9 @@ from mockintosh.params import (
     BodyUrlencodedParam,
     BodyMultipartParam
 )
-from mockintosh.templating import TemplateRenderer
 from mockintosh.stats import Stats
+from mockintosh.logs import Logs, LogRecord
+from mockintosh.templating import TemplateRenderer
 
 OPTIONS = 'options'
 ORIGIN = 'Origin'
@@ -72,6 +74,8 @@ IMAGE_EXTENSIONS = [
     '.webp'
 ]
 
+FALLBACK_TO_TIMEOUT = int(os.environ.get('MOCKINTOSH_FALLBACK_TO_TIMEOUT', 30))
+
 hbs_random = hbs_Random()
 j2_random = j2_Random()
 
@@ -80,6 +84,8 @@ j2_date = j2_Date()
 
 counters = {}
 
+client = httpx.AsyncClient()
+
 __location__ = os.path.abspath(os.path.dirname(__file__))
 
 
@@ -88,62 +94,21 @@ class NewHTTPError(Exception):
     pass
 
 
-class NotParsedJSON():
-    """Class to determine wheter the request body is parsed into JSON or not."""
-    pass
-
-
-class Request():
-    """Class that defines the `Request` object which is being injected into the response template."""
-
-    def __init__(self) -> None:
-        self.version = None
-        self.remoteIp = None
-        self.protocol = None
-        self.host = None
-        self.hostName = None
-        self.uri = None
-        self.method = None
-        self.path = None
-        self.headers = {}
-        self.queryString = {}
-        self.body = {}
-        self._json = NotParsedJSON()
-
-    @property
-    def json(self) -> [None, dict]:
-        if isinstance(self._json, NotParsedJSON):
-            try:
-                self._json = json.loads(self.body)
-            except json.JSONDecodeError:
-                logging.warning('Failed to decode request body to JSON:\n%s', self.body)
-                self._json = None
-        return self._json
-
-
-class Response():
-    """Class that defines the `Response` object which is being used by the interceptors."""
-
-    def __init__(self) -> None:
-        self.status = None
-        self.headers = {}
-        self.body = None
-
-
 class GenericHandler(tornado.web.RequestHandler):
     """Class to handle all mocked requests."""
 
     def prepare(self) -> Optional[Awaitable[None]]:
         """Overriden method of tornado.web.RequestHandler"""
         self.dont_add_status_code = False
-        self.special_request_start_time = time.perf_counter()
         super().prepare()
 
     def on_finish(self) -> None:
         """Overriden method of tornado.web.RequestHandler"""
+        elapsed_time = self.request.request_time()
+        self.add_log_record(int(round(elapsed_time * 1000)))
         if self.get_status() != 500 and not self.is_options and self.methods is not None:
             if self.get_status() != 405:
-                self.set_elapsed_time()
+                self.set_elapsed_time(elapsed_time)
             if not self.dont_add_status_code:
                 if self.get_status() == 405:
                     self.stats.services[self.service_id].add_status_code(
@@ -155,15 +120,37 @@ class GenericHandler(tornado.web.RequestHandler):
                     )
         super().on_finish()
 
-    def set_elapsed_time(self) -> None:
+    def write(self, chunk: Union[str, bytes, dict]) -> None:
+        super().write(chunk)
+        if hasattr(self, 'replica_response'):
+            self.replica_response.bodySize = len(b"".join(self._write_buffer))
+
+    def set_elapsed_time(self, elapsed_time_in_seconds: float) -> None:
         """Method to calculate and store the elapsed time of the request handling to be used in stats."""
-        elapsed_time_in_seconds = (time.perf_counter() - self.special_request_start_time) / 1000000
         self.stats.services[self.service_id].endpoints[self.internal_endpoint_id].add_request_elapsed_time(
             elapsed_time_in_seconds
         )
 
+    def add_log_record(self, elapsed_time_in_milliseconds: int) -> None:
+        """Method that creates a log record and inserts it to log tracking system."""
+        if not self.logs.services[self.service_id].is_enabled() or self.request.server_connection.stream.socket is None:
+            return
+
+        request_start_datetime = datetime.fromtimestamp(self.request._start_time)
+        request_start_datetime.replace(tzinfo=timezone.utc)
+        log_record = LogRecord(
+            self.logs.services[self.service_id].name,
+            request_start_datetime,
+            elapsed_time_in_milliseconds,
+            self.replica_request,
+            self.replica_response,
+            self.request.server_connection
+        )
+        self.logs.services[self.service_id].add_record(log_record)
+
     def initialize(
         self,
+        http_server,
         config_dir: str,
         service_id: int,
         endpoints: list,
@@ -171,19 +158,24 @@ class GenericHandler(tornado.web.RequestHandler):
         definition_engine: str,
         interceptors: list,
         stats: Stats,
+        logs: Logs,
         unhandled_data,
-        tag: str
+        fallback_to: Union[str, None],
+        tag: Union[str, None]
     ) -> None:
         """Overriden method of tornado.web.RequestHandler"""
         try:
+            self.http_server = http_server
             self.config_dir = config_dir
             self.endpoints = endpoints
             self.methods = None
             self.custom_args = ()
             self.stats = stats
+            self.logs = logs
             self.service_id = service_id
             self.internal_endpoint_id = None
             self.unhandled_data = unhandled_data
+            self.fallback_to = fallback_to
             self.tag = tag
 
             for path, methods in self.endpoints:
@@ -203,29 +195,30 @@ class GenericHandler(tornado.web.RequestHandler):
             self.is_options = False
             self.custom_dataset = {}
 
-            self.special_request = self.build_special_request()
+            self.replica_request = self.build_replica_request()
             self.default_context = {
-                'request': self.special_request
+                'request': self.replica_request
             }
             self.custom_context = {}
-        except Exception as e:
+        except Exception as e:  # pragma: no cover
             self.set_status(500)
             self.write(''.join(traceback.format_tb(e.__traceback__)))
             self.write('%s' % str(e))
 
-    def super_verb(self, *args) -> None:
+    async def super_verb(self, *args) -> None:
         """A method to unify all the HTTP verbs under a single flow."""
         try:
+            self.args_backup = args
             self.set_default_headers()
 
             if not self.is_options:
                 if self.custom_args:
                     args = self.custom_args
                 if self.methods is None:
-                    self.raise_http_error(404)
-                self.dynamic_unimplemented_method_guard()
+                    await self.raise_http_error(404)
+                await self.dynamic_unimplemented_method_guard()
 
-            match_alternative_return = self.match_alternative()
+            match_alternative_return = await self.match_alternative()
             if not match_alternative_return:
                 return
             _id, response, params, context, dataset, internal_endpoint_id, performance_profile = match_alternative_return
@@ -247,44 +240,44 @@ class GenericHandler(tornado.web.RequestHandler):
 
             if self.rendered_body is None:
                 return
-            self.special_response = self.build_special_response()
+            self.replica_response = self.build_replica_response()
             if self.should_write():
                 self.write(self.rendered_body)
         except NewHTTPError:
             return
-        except Exception as e:
+        except Exception as e:  # pragma: no cover
             self.set_status(500)
             self.write(''.join(traceback.format_tb(e.__traceback__)))
             self.write('%s' % str(e))
 
-    def get(self, *args) -> None:
+    async def get(self, *args) -> None:
         """Overriden method of tornado.web.RequestHandler"""
-        self.super_verb(*args)
+        await self.super_verb(*args)
 
-    def post(self, *args) -> None:
+    async def post(self, *args) -> None:
         """Overriden method of tornado.web.RequestHandler"""
-        self.super_verb(*args)
+        await self.super_verb(*args)
 
-    def head(self, *args) -> None:
+    async def head(self, *args) -> None:
         """Overriden method of tornado.web.RequestHandler"""
-        self.super_verb(*args)
+        await self.super_verb(*args)
 
-    def delete(self, *args) -> None:
+    async def delete(self, *args) -> None:
         """Overriden method of tornado.web.RequestHandler"""
-        self.super_verb(*args)
+        await self.super_verb(*args)
 
-    def patch(self, *args) -> None:
+    async def patch(self, *args) -> None:
         """Overriden method of tornado.web.RequestHandler"""
-        self.super_verb(*args)
+        await self.super_verb(*args)
 
-    def put(self, *args) -> None:
+    async def put(self, *args) -> None:
         """Overriden method of tornado.web.RequestHandler"""
-        self.super_verb(*args)
+        await self.super_verb(*args)
 
-    def options(self, *args) -> None:
+    async def options(self, *args) -> None:
         """Overriden method of tornado.web.RequestHandler"""
         self.is_options = True
-        self.super_verb(*args)
+        await self.super_verb(*args)
 
     def populate_context(self, *args) -> None:
         """Method to populate the context to be used by the templating engines."""
@@ -292,11 +285,8 @@ class GenericHandler(tornado.web.RequestHandler):
         for key, value in self.custom_dataset.items():
             self.custom_context[key] = value
         if args:
-            if len(args) >= len(self.initial_context):
-                for i, key in enumerate(self.initial_context):
-                    self.custom_context[key] = args[i]
-            else:
-                self.raise_http_error(404)
+            for i, key in enumerate(self.initial_context):
+                self.custom_context[key] = args[i]
         self.custom_context.update(self.default_context)
         self.analyze_component('headers')
         self.analyze_component('queryString')
@@ -307,20 +297,15 @@ class GenericHandler(tornado.web.RequestHandler):
 
     def populate_counters(self, context: [None, dict]) -> None:
         """Method that retrieves counters from template engine contexts."""
-        if context is None:
-            return
-
         if SPECIAL_CONTEXT in context and 'counters' in context[SPECIAL_CONTEXT]:
             for key, value in context[SPECIAL_CONTEXT]['counters'].items():
                 counters[key] = value
 
-    def dynamic_unimplemented_method_guard(self) -> None:
+    async def dynamic_unimplemented_method_guard(self) -> None:
         """Method to handle unimplemented HTTP verbs (`405`)."""
-        if self.methods is None:
-            self.raise_http_error(404)
         if self.request.method.lower() not in self.methods:
             self.write('Supported HTTP methods: %s' % ', '.join([x.upper() for x in self.methods.keys()]))
-            self.raise_http_error(405)
+            await self.raise_http_error(405)
 
     def log_request(self) -> None:
         """Method that logs the request."""
@@ -338,11 +323,11 @@ class GenericHandler(tornado.web.RequestHandler):
             if isinstance(param, QueryStringParam):
                 context[key] = self.get_query_argument(param.key)
             if isinstance(param, BodyTextParam):
-                context[key] = _decoder(self.request.body)
+                context[key] = self.request.body.decode()
             if isinstance(param, BodyUrlencodedParam):
                 context[key] = self.get_body_argument(param.key)
             if isinstance(param, BodyMultipartParam):
-                context[key] = _decoder(self.request.files[param.key][0].body)
+                context[key] = self.request.files[param.key][0].body.decode()
         return context
 
     def render_template(self) -> str:
@@ -362,7 +347,7 @@ class GenericHandler(tornado.web.RequestHandler):
                 logging.debug('Reading external file from path: %s', template_path)
                 source_text = file.read()
                 try:
-                    source_text = source_text.decode('utf-8')
+                    source_text = source_text.decode()
                     logging.debug('Template file text: %s', source_text)
                 except UnicodeDecodeError:
                     is_binary = True
@@ -381,7 +366,7 @@ class GenericHandler(tornado.web.RequestHandler):
 
         return compiled
 
-    def build_special_request(self) -> Request:
+    def build_replica_request(self) -> Request:
         """Method that builds the `Request` object to be injected into the response templating."""
         request = Request()
 
@@ -391,6 +376,7 @@ class GenericHandler(tornado.web.RequestHandler):
         request.protocol = self.request.protocol
         request.host = self.request.host
         request.hostName = self.request.host_name
+        request.port = self.request.server_connection.stream.socket.getsockname()[1]
         request.uri = self.request.uri
 
         # Method
@@ -406,29 +392,48 @@ class GenericHandler(tornado.web.RequestHandler):
 
         # Query String
         for key, value in self.request.query_arguments.items():
-            request.queryString[key] = [_decoder(x) for x in value]
+            request.queryString[key] = [x.decode() for x in value]
             if len(request.queryString[key]) == 1:
                 request.queryString[key] = request.queryString[key][0]
 
         # Body
         if self.request.body_arguments:
+            request.mimeType = 'application/x-www-form-urlencoded'
             for key, value in self.request.body_arguments.items():
-                request.body[key] = [_decoder(x) for x in value]
+                try:
+                    request.bodyType[key] = 'str'
+                    request.body[key] = [x.decode() for x in value]
+                except (AttributeError, UnicodeDecodeError):
+                    request.bodyType[key] = BASE64
+                    request.body[key] = [_b64encode(x) for x in value]
                 if len(request.body[key]) == 1:
                     request.body[key] = request.body[key][0]
         elif self.request.files:
+            request.mimeType = 'multipart/form-data'
             for key, value in self.request.files.items():
-                request.body[key] = [_decoder(x.body) for x in value]
+                try:
+                    request.bodyType[key] = 'str'
+                    request.body[key] = [x.body.decode() for x in value]
+                except (AttributeError, UnicodeDecodeError):
+                    request.bodyType[key] = BASE64
+                    request.body[key] = [_b64encode(x.body) for x in value]
                 if len(request.body[key]) == 1:
                     request.body[key] = request.body[key][0]
         else:
-            request.body = _decoder(self.request.body)
+            request.mimeType = 'text/plain'
+            try:
+                request.bodyType = 'str'
+                request.body = self.request.body.decode()
+            except (AttributeError, UnicodeDecodeError):
+                request.bodyType = BASE64
+                request.body = _b64encode(self.request.body)
+        request.bodySize = len(self.request.body)
 
         request.files = self.request.files
 
         return request
 
-    def build_special_response(self) -> Response:
+    def build_replica_response(self) -> Response:
         """Method that prepares `Response` object to be modified by the interceptors."""
         response = Response()
 
@@ -442,9 +447,9 @@ class GenericHandler(tornado.web.RequestHandler):
 
     def update_response(self) -> None:
         """Updates the response according to modifications made in interceptors."""
-        self._status_code = self.special_response.status
-        self._headers = self.special_response.headers
-        self.rendered_body = self.special_response.body
+        self._status_code = self.replica_response.status
+        self._headers = self.replica_response.headers
+        self.rendered_body = self.replica_response.body
         self._write_buffer = []
         if self.rendered_body is None:
             self.rendered_body = ''
@@ -480,7 +485,7 @@ class GenericHandler(tornado.web.RequestHandler):
                 struct.pack('ii', 1, 0)
             )
             self.request.server_connection.stream.close()
-            self.set_elapsed_time()
+            self.set_elapsed_time(self.request.request_time())
             self.stats.services[self.service_id].endpoints[self.internal_endpoint_id].add_status_code('RST')
         if isinstance(status_code, str) and status_code.lower() == 'fin':
             self.request.server_connection.stream.close()
@@ -500,7 +505,7 @@ class GenericHandler(tornado.web.RequestHandler):
         elif component == 'queryString':
             payload = self.request.query_arguments
         elif component == 'bodyText':
-            payload = _decoder(self.request.body)
+            payload = self.request.body.decode()
         elif component == 'bodyUrlencoded':
             payload = self.request.body_arguments
         elif component == 'bodyMultipart':
@@ -522,7 +527,7 @@ class GenericHandler(tornado.web.RequestHandler):
                     elif component == 'bodyUrlencoded':
                         match_string = self.get_body_argument(key)
                     elif component == 'bodyMultipart':
-                        match_string = _decoder(self.request.files[key][0].body)
+                        match_string = self.request.files[key][0].body.decode()
 
                     match = re.search(value['regex'], match_string)
                     if match is not None:
@@ -563,24 +568,22 @@ class GenericHandler(tornado.web.RequestHandler):
             for value in new_value_list:
                 if key.title() == 'Set-Cookie':
                     value_splitted = value.split('=')
-                    value_splitted[1] = urllib.parse.quote_plus(value_splitted[1])
+                    value_splitted[1] = quote_plus(value_splitted[1])
                     self.set_cookie(value_splitted[0], value_splitted[1])
                 else:
                     self.set_header(key, value)
 
-    def match_alternative(self) -> tuple:
+    async def match_alternative(self) -> tuple:
         """Method to handles all the request matching logic.
 
         If the request does not match to any alternatives defined in the config, it returns `400`.
 
         It also handles the automatic CORS.
         """
-        if self.should_cors():
+        if await self.should_cors():
             self.respond_cors()
             return ()
 
-        if self.methods is None:
-            return ()
         self.alternatives = self.methods[self.request.method.lower()]
 
         response = None
@@ -626,11 +629,6 @@ class GenericHandler(tornado.web.RequestHandler):
                         fail = True
                         reason = 'Key \'%s\' couldn\'t found in the query string!' % key
                         break
-                    if key not in self.request.query_arguments:
-                        self.internal_endpoint_id = alternative['internalEndpointId']
-                        fail = True
-                        reason = 'Key \'%s\' couldn\'t found in the query string!' % key
-                        break
                     if value == request_query_val:
                         continue
                     value = '^%s$' % value
@@ -649,7 +647,7 @@ class GenericHandler(tornado.web.RequestHandler):
 
             # Body
             if 'body' in alternative:
-                body = _decoder(self.request.body)
+                body = self.request.body.decode()
 
                 # Schema
                 if 'schema' in alternative['body']:
@@ -658,7 +656,14 @@ class GenericHandler(tornado.web.RequestHandler):
                         json_schema_path = self.resolve_relative_path(json_schema)
                         with open(json_schema_path, 'r') as file:
                             logging.info('Reading JSON schema file from path: %s', json_schema_path)
-                            json_schema = json.load(file)
+                            try:
+                                json_schema = json.load(file)
+                            except json.decoder.JSONDecodeError:
+                                self.send_error(
+                                    500,
+                                    message='JSON decode error of the JSON schema file: %s' % json_schema
+                                )
+                                return
                             logging.debug('JSON schema: %s', json_schema)
                     json_data = None
 
@@ -694,7 +699,7 @@ class GenericHandler(tornado.web.RequestHandler):
                             reason = 'Request body:\n\n%s\nDeos not match to regex:\n\n%s' % (body, value)
                             break
 
-                # Urlncoded
+                # Urlencoded
                 if 'urlencoded' in alternative['body']:
                     for key, value in alternative['body']['urlencoded'].items():
                         # To prevent 400, default=None
@@ -705,11 +710,6 @@ class GenericHandler(tornado.web.RequestHandler):
                             fail = True
                             reason = 'Key \'%s\' couldn\'t found in the form data!' % key
                             break
-                        if key not in self.request.body_arguments:
-                            self.internal_endpoint_id = alternative['internalEndpointId']
-                            fail = True
-                            reason = 'Key \'%s\' couldn\'t found in the form data!' % key
-                            break
                         if value == body_argument:
                             continue
                         value = '^%s$' % value
@@ -734,7 +734,7 @@ class GenericHandler(tornado.web.RequestHandler):
                             fail = True
                             reason = 'Key \'%s\' couldn\'t found in the multipart data!' % key
                             break
-                        multipart_argument = _decoder(self.request.files[key][0].body)
+                        multipart_argument = self.request.files[key][0].body.decode()
                         if value == multipart_argument:
                             continue
                         value = '^%s$' % value
@@ -751,29 +751,20 @@ class GenericHandler(tornado.web.RequestHandler):
                     if fail:
                         continue
 
-            if self.should_cors():
-                self.respond_cors()
-
             # Multiple responses
             if 'response' in alternative:
                 response = alternative['response']
                 if isinstance(alternative['response'], list):
                     if not len(alternative['response']) > 0:
-                        response = {
-                            'body': None
-                        }
+                        response = {'body': None}
                     else:
                         response = self.loop_alternative(alternative, 'response', 'multiResponses')
                         if not response:
                             return ()
 
-                response = response if isinstance(response, dict) else {
-                    'body': response
-                }
+                response = response if isinstance(response, dict) else {'body': response}
             else:
-                response = {
-                    'body': None
-                }
+                response = {'body': None}
 
             # Dataset
             dataset = {}
@@ -792,19 +783,18 @@ class GenericHandler(tornado.web.RequestHandler):
             return (_id, response, params, context, dataset, internal_endpoint_id, performance_profile)
 
         self.write(reason)
-        self.raise_http_error(400)
-        self.finish()
+        await self.raise_http_error(400)
 
     def trigger_interceptors(self) -> None:
         """Method to trigger the interceptors"""
         for interceptor in self.interceptors:
-            interceptor(self.special_request, self.special_response)
+            interceptor(self.replica_request, self.replica_response)
 
     def finish(self, chunk: Optional[Union[str, bytes, dict]] = None) -> "Future[None]":
         """Overriden method of tornado.web.RequestHandler"""
         if self._status_code not in (204, 500, 'RST', 'FIN'):
-            if not hasattr(self, 'special_response'):
-                self.special_response = self.build_special_response()
+            if not hasattr(self, 'replica_response'):
+                self.replica_response = self.build_replica_response()
             self.trigger_interceptors()
             if self.interceptors:
                 self.update_response()
@@ -828,16 +818,16 @@ class GenericHandler(tornado.web.RequestHandler):
         orig_relative_path, context = self.common_template_renderer(self.definition_engine, orig_relative_path)
         self.populate_counters(context)
 
-        error_msg = 'External template file \'%s\' couldn\'t be accessed or found!' % orig_relative_path
         if orig_relative_path[0] == '/':
             orig_relative_path = orig_relative_path[1:]
+        error_msg = 'External template file \'%s\' couldn\'t be accessed or found!' % orig_relative_path
         relative_path = os.path.join(self.config_dir, orig_relative_path)
         if not os.path.isfile(relative_path):
-            self.send_error(403, message=error_msg)
+            self.send_error(500, message=error_msg)
             return None
         relative_path = os.path.abspath(relative_path)
         if not relative_path.startswith(self.config_dir):
-            self.send_error(403, message=error_msg)
+            self.send_error(500, message=error_msg)
             return None
 
         return relative_path
@@ -846,7 +836,7 @@ class GenericHandler(tornado.web.RequestHandler):
         """Overriden method of tornado.web.RequestHandler"""
         if 'message' in kwargs and kwargs['message']:
             self.finish(kwargs['message'])
-        else:
+        else:  # pragma: no cover
             self.finish()
 
     def respond_cors(self) -> None:
@@ -862,13 +852,13 @@ class GenericHandler(tornado.web.RequestHandler):
     def set_cors_headers(self) -> None:
         """Method that sets the CORS headers."""
         if ORIGIN in self.request.headers._dict:
-            self.set_header('access-control-allow-methods', 'DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT')
+            self.set_header('Access-Control-Allow-Methods', 'DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT')
             origin = self.request.headers.get(ORIGIN)
-            self.set_header('access-control-allow-origin', origin)
+            self.set_header('Access-Control-Allow-Origin', origin)
 
             if AC_REQUEST_HEADERS in self.request.headers._dict:
                 ac_request_headers = self.request.headers.get(AC_REQUEST_HEADERS)
-                self.set_header('access-control-allow-headers', ac_request_headers)
+                self.set_header('Access-Control-Allow-Headers', ac_request_headers)
 
     def set_default_headers(self) -> None:
         """Method that sets the default headers."""
@@ -876,12 +866,13 @@ class GenericHandler(tornado.web.RequestHandler):
             PROGRAM.capitalize(),
             mockintosh.__version__
         ))
+        self.set_header('X-Mockintosh-Prompt', "Hello, I'm Mockintosh.")  # clear signature that it's mock
         self.set_cors_headers()
 
-    def should_cors(self) -> bool:
+    async def should_cors(self) -> bool:
         """Method that decides whether the request is applicable for automatic CORS or not."""
         if self.is_options and self.methods is None:
-            self.raise_http_error(404)
+            await self.raise_http_error(404)
         return self.is_options and self.methods is not None and self.request.method.lower() not in self.methods.keys()
 
     def load_dataset(self, dataset: [list, str]) -> dict:
@@ -917,15 +908,7 @@ class GenericHandler(tornado.web.RequestHandler):
                 return False
 
         if 'tag' in alternative[key][alternative[index_key]]:
-            if self.tag is None:
-                if resetted:
-                    self.internal_endpoint_id = alternative['internalEndpointId']
-                    self.set_status(410)
-                    self.finish()
-                    return False
-                else:
-                    return self.loop_alternative(alternative, key, subkey)
-            elif self.tag != alternative[key][alternative[index_key]]['tag']:
+            if self.tag is None or not self.tag or self.tag != alternative[key][alternative[index_key]]['tag']:
                 if resetted:
                     self.internal_endpoint_id = alternative['internalEndpointId']
                     self.set_status(410)
@@ -948,8 +931,6 @@ class GenericHandler(tornado.web.RequestHandler):
             from mockintosh.j2.methods import fake, counter, json_path, escape_html
             self.custom_context['random'] = j2_random
             self.custom_context['date'] = j2_date
-        else:
-            raise UnsupportedTemplateEngine(template_engine, SUPPORTED_ENGINES)
 
         renderer = TemplateRenderer(
             template_engine,
@@ -966,13 +947,9 @@ class GenericHandler(tornado.web.RequestHandler):
         )
         return renderer.render()
 
-    def raise_http_error(self, status_code: int) -> None:
+    async def raise_http_error(self, status_code: int) -> None:
         """Method to throw a `NewHTTPError`."""
-        if self.unhandled_data is not None:
-            identifier = '%s %s' % (self.request.method.upper(), self.request.path)
-            if identifier not in self.unhandled_data.requests[self.service_id]:
-                self.unhandled_data.requests[self.service_id][identifier] = []
-            self.unhandled_data.requests[self.service_id][identifier].append(self.request)
+        await self.resolve_unhandled_request()
 
         self.set_status(status_code)
 
@@ -988,3 +965,105 @@ class GenericHandler(tornado.web.RequestHandler):
                     self.rendered_body = image
 
         raise NewHTTPError()
+
+    async def resolve_unhandled_request(self) -> None:
+        if self.fallback_to is None:
+            self.insert_unhandled_data((self.request, None))
+            return
+
+        # Headers
+        headers = {}
+        for key, value in self.request.headers._dict.items():
+            if key.title() in (
+                'Host',
+                'Content-Length'
+            ):
+                continue
+            headers[key] = value
+        headers['Cache-Control'] = 'no-cache'
+        headers['If-None-Match'] = '0'
+
+        # Query String
+        query_string = ''
+        for key, value in self.request.query_arguments.items():
+            if not query_string:
+                query_string = '?'
+            values = [x.decode() for x in value]
+            if len(values) == 1:
+                query_string += '%s=%s' % (key, values[0])
+            else:
+                for _value in values:
+                    query_string += '%s[]=%s' % (key, _value)
+
+        # Body
+        data = {}
+        files = {}
+        if self.request.body_arguments:
+            for key, value in self.request.body_arguments.items():
+                try:
+                    data[key] = [x.decode() for x in value]
+                except (AttributeError, UnicodeDecodeError):
+                    data[key] = [x for x in value]
+                if len(data[key]) == 1:
+                    data[key] = data[key][0]
+        elif self.request.files:
+            for key, value in self.request.files.items():
+                try:
+                    files[key] = [x.body.decode() for x in value]
+                except (AttributeError, UnicodeDecodeError):
+                    files[key] = [x.body for x in value]
+                if len(files[key]) == 1:
+                    files[key] = files[key][0]
+        else:
+            data = self.request.body.decode()
+
+        url = self.fallback_to.rstrip('/') + self.request.path + query_string
+
+        # The service is external
+        logging.info('Redirecting the unhandled request to: %s %s' % (self.request.method, url))
+
+        http_verb = getattr(client, self.request.method.lower())
+        try:
+            if self.request.method.upper() in ('POST', 'PUT', 'PATCH', 'DELETE'):
+                resp = await http_verb(url, headers=headers, timeout=FALLBACK_TO_TIMEOUT, data=data, files=files)
+            else:
+                resp = await http_verb(url, headers=headers, timeout=FALLBACK_TO_TIMEOUT)
+        except httpx.TimeoutException:
+            self.set_status(504)
+            self.write('Redirected request to: %s %s is timed out!' % (self.request.method, url))
+            raise NewHTTPError()
+        except httpx.ConnectError:
+            self.set_status(502)
+            self.write('Name or service not known: %s' % self.fallback_to.rstrip('/'))
+            raise NewHTTPError()
+
+        logging.info('Returned back from the redirected request.')
+
+        self.set_status(resp.status_code if resp.status_code != 304 else 200)
+        for key, value in resp.headers.items():
+            if key.title() in (
+                'Server',
+                'Transfer-Encoding',
+                'Content-Length',
+                'Content-Encoding',
+                'Access-Control-Allow-Methods',
+                'Access-Control-Allow-Origin'
+            ):
+                continue
+            self.set_header(key, value)
+
+        self.write(resp.content)
+        self.replica_response = self.build_replica_response()
+        self.replica_response.body = resp.content
+
+        self.insert_unhandled_data((self.request, self.replica_response))
+        raise NewHTTPError()
+
+    def insert_unhandled_data(self, row: tuple) -> None:
+        if self.unhandled_data is None:
+            return
+
+        identifier = '%s %s' % (self.request.method.upper(), self.request.path)
+        if identifier not in self.unhandled_data.requests[self.service_id]:
+            self.unhandled_data.requests[self.service_id][identifier] = []
+        self.unhandled_data.requests[self.service_id][identifier].append(row)
